#ifndef _CONFIDENCE_MEASURES
#define _CONFIDENCE_MEASURES
#define _CRT_SECURE_NO_WARNINGS

#include<stdio.h>
#include<stdlib.h>
#include <string.h>
#include <iostream>
#include <fcntl.h>
#include <fstream>

#include "confidence_measures_utility.h"

#include<opencv/cv.h>
#include<opencv/highgui.h>

using namespace cv;

struct confParams
{
	bool useMSM;
	bool useLC;
	bool useCUR;
	bool usePKR;
	bool usePKRN;
	bool useAPKR;
	bool useAPKRN;
	bool useMM;
	bool useMMN;
	bool useNMM;
	bool useNMMN;
	bool useDVM;
	bool useDAM;
	bool useMLM;
	bool useAML;
	bool useNEM;
	bool useNOI;
	bool useWMN;
	bool useWMNN;
	bool usePER;
	bool useLRC;
	bool useACC;
	bool useLRD;
	bool useHGM;
	bool useZSAD;
	bool useDTS;
	bool useDSM;
	bool useSAMM;
	bool useDB;
	bool useDBL;
	bool useDTD;
	bool useDTE;
	bool useUC;
	bool useMED;
	bool useVAR;
	bool useLMN;
	bool useWPKR;
	bool useWPKRN;

	int zsad_radius,
		apkr_radius,
		lmn_radius,
		med_radius,
		DD_edge_threshold,
		DD_ratio,
		DD_radius,
		wpkr_radius,
		DD_max_lowThreshold;


	float pkr_epsilon,
		apkr_epsilon,
		dsm_epsilon,
		lrd_epsilon,
		mlm_sigma,
		aml_sigma,
		per_sigma,
		nmm_sigma,
		wpkr_epsilon,
		wpkr_threshold,
		lc_gamma;

	confParams() :  useMSM(false),
					useLC(false),
					useCUR(false),
					usePKR(false),
					usePKRN(false),
					useAPKR(false),
					useAPKRN(false),
					useMM(false),
					useMMN(false),
					useNMM(false),
					useNMMN(false),
					useDVM(false),
					useDAM(false),
					useMLM(false),
					useAML(false),
					useNEM(false),
					useNOI(false),
					useWMN(false),
					useWMNN(false),
					usePER(false),
					useLRC(false),
					useACC(false),
					useLRD(false),
					useHGM(false),
					useZSAD(false),
					useDTS(false),
					useDSM(false),
					useSAMM(false),
					useDB(false),
					useDBL(false),
					useDTD(false),
					useDTE(false),
					useUC(false),
					useMED(false),
					useLMN(false),
					useWPKR(false),
					useWPKRN(false),
	
					zsad_radius(-1),
					apkr_radius(-1),
					lmn_radius(-1),
					med_radius(-1),
					pkr_epsilon(-1),
					apkr_epsilon(-1),
					dsm_epsilon(-1),
					lrd_epsilon(-1),
					mlm_sigma(-1),
					aml_sigma(-1),
					per_sigma(-1),
					nmm_sigma(-1),
					lc_gamma(-1),
					DD_edge_threshold(-1),
					DD_ratio(-1),
					wpkr_radius(-1),
					wpkr_epsilon(-1),
					wpkr_threshold(-1),
					DD_radius(-1) {}
};

/***************************************************************************************************/
/*
* Confidence Measures
*
/***************************************************************************************************/

//	3.1 Minimum cost and local properties of the cost curve  
void matching_score_measure(InputArray c1, int height, int width, OutputArray confidence_map);
void maximum_margin(InputArray c1, InputArray c2m, int height, int width, OutputArray confidence_map);
void non_linear_margin(InputArray c1, InputArray c2m, float sigma, int height, int width, OutputArray confidence_map);
void curvature(InputArrayOfArrays costs, InputArray c1_idx, int height, int width, OutputArray confidence_map);
void local_curve(InputArrayOfArrays costs, InputArray c1_idx, float gamma, int height, int width, OutputArray confidence_map);
void peak_ratio(InputArray c1, InputArray c2m, float epsilon, int height, int width, OutputArray confidence_map);
void average_peak_ratio(InputArrayOfArrays costs, InputArray c1_idx, InputArray c2m_idx, float epsilon, int r, int height, int width, OutputArray confidence_map);
void weighted_peak_ratio(InputArrayOfArrays costs, InputArray left, InputArray c1_idx, InputArray c2m_idx, float epsilon, float threshold_color, int r, int height, int width, OutputArray confidence_map);
void disparity_ambiguity_measure(InputArray c1_idx, InputArray c2_idx, int height, int width, OutputArray confidence_map);
void semi_global_energy(InputArray c1, InputArray disparity_L2R, int p1, int p2, int r, int height, int width, OutputArray confidence_map);

//	3.2 Analysis of the entire cost curve
void perturbation_measure(InputArrayOfArrays costs, InputArray c1, InputArray c1_idx, float s, int height, int width, OutputArray confidence_map);
void maximum_likelihood_measure(InputArrayOfArrays costs, InputArray c1, float sigma, int height, int width, OutputArray confidence_map);
void attainable_maximum_likelihood(InputArrayOfArrays costs, InputArray c1, float sigma, int height, int width, OutputArray confidence_map);
// 	NOI: just run inflections function
void local_minima_in_neighborhood(InputArrayOfArrays local_minima, InputArray c1_idx, int r, int height, int width, OutputArray confidence_map);
void winner_margin(InputArray c1, InputArray c2m, InputArray c_sum, int height, int width, OutputArray confidence_map);
void negative_entropy_measure(InputArrayOfArrays costs, int height, int width, OutputArray confidence_map);

//	3.3 Left and right consistency
void left_right_consistency_check(InputArray disparity_L2R, InputArray disparity_R2L, int disp_scale, int height, int width, OutputArray confidence_map);
void left_right_difference(InputArray c1_L, InputArray c2_L, InputArray c1_R, InputArray disparity_L, int disp_scale, float epsilon, int height, int width, OutputArray confidence_map);
void zero_mean_sum_of_absolute_differences(InputArray left, InputArray right, InputArray c1, InputArray c1_idx, int r, int height, int width, OutputArray confidence_map);
void uniqueness_constraint(InputArray c1, InputArray c1_idx, int height, int width, OutputArray confidence_map);
void asymmetric_consistency_check(InputArray c1, InputArray c1_idx, int height, int width, OutputArray confidence_map);

//	3.4 Disparity map features
void compute_DD(InputArray image, int lowThreshold, int ratio, int r, OutputArray confidence_map);
void disparity_variance_measure(InputArray disparity_map, int size, OutputArray confidence_map);
void var_measure(InputArray disparity_map, int r, OutputArray confidence_map);
void difference_with_median(InputArray disparity_L2R, int r, int height, int width, OutputArray confidence_map);
//	you can find DA and DS here: http://vision.disi.unibo.it/~mpoggi/code/3DV2016_O1.zip

//	3.5 Reference image features
void distance_to_border(InputArray image, OutputArray confidence_map);
void distance_to_left_border(InputArray image, int d_max, OutputArray confidence_map);
void horizontal_gradient(InputArray left_stereo, OutputArray confidence_map);
//	DTE: use compute_DD with reference image as input

//	3.6 Image distinctiveness
void distinctiveness(InputArrayOfArrays costs_LL, int d_min, int d_max, int height, int width, OutputArray confidence_map);
void distinctive_similarity_measure(InputArray distinctiveness_L, InputArray distinctiveness_R, InputArray c1, InputArray c1_idx, float epsilon, int height, int width, OutputArray confidence_map);
void self_aware_matching(InputArrayOfArrays costs, InputArrayOfArrays costs_LL, InputArray c1_idx, int d_min, int d_max, int height, int width, OutputArray confidence_map);

//	3.8 SGM specific
void PS_measure(Mat cl1, Mat cl2, Mat cl1_idx, Mat cl2_idx, Mat cg1_idx, int height, int width, float gamma, OutputArray confidence_map);
void sum_of_consistent_disparites(InputArray disparity_map, InputArrayOfArrays scanline_disparity_maps, int height, int width, OutputArray confidence_map);

//	More functions:

//	Compute the error to ground-truth
void error_to_groundtruth(InputArray disparity_L2R, InputArray groundtruth, int height, int width, OutputArray confidence_map);

//	Sample functions processing (most of the) confidence measures
void fn_confidence_measure(InputArray image_L, InputArray image_R, _DSI dsi_LR, _DSI dsi_LL, _DSI dsi_RR, confParams params, OutputArray disparity_map, vector<string> &methods, OutputArrayOfArrays confidences);

#endif 
