require 'torch'   -- torch
require 'image'   -- for color transforms
require 'gfx.js'  -- to visualize the dataset
require 'nn'      -- provides a normalization operator
require 'xlua'	  -- useful tools (progress bar)


----------------------------------------------------------------------
-- parse command line arguments
if not opt then
   print '==> processing options'
   cmd = torch.CmdLine()
   cmd:text()
   cmd:text('Dataset Preprocessing')
   cmd:text()
   cmd:text('Options:')
   cmd:option('-inputFolder', '/home/titan/Scrivania/ICCV_SURVEY_CODE/ad/KITTI/', 'Input folder')
   cmd:option('-first', 0, 'First image')
   cmd:option('-last', 24, 'Last image')
   cmd:option('-disparityName', 'disparity_LR.png', 'File name for disparity maps')
   cmd:option('-d1Name', 'p1.png', 'File name for disparity maps')
   cmd:option('-d2Name', 'p2.png', 'File name for disparity D2')
   cmd:option('-datasetFolder', './DATASET_ad/', 'dataset folder')
   cmd:option('-gtFolder', '/KITTI/disp_noc/', 'dataset folder')
   cmd:option('-leftFolder', '/KITTI/image_0/', 'dataset folder')
   cmd:option('-size', 9, 'patch size')
   cmd:option('-createDataset', false, 'Creating dataset')
   cmd:option('-mode', 'CCNN', 'Training mode')
   cmd:option('-dataset', 'kitti', 'Training dataset')

   cmd:text()
   opt = cmd:parse(arg or {})
end
----------------------------------------------------------------------

torch.setdefaulttensortype('torch.FloatTensor') -- preprocessing requires a floating point representation 


function create_PBCP_dataset(first, last)

	local dataset = {}
	local tot_samples = 0
	local samples = {}
	local labels = {}
	local valids = {}
	--local corrects = 0;

	-- loading images 	
	for t = first, last do

		local image_id    
		if t < 10 then
			image_id = '00000'..t
		else
			image_id = '0000'..t
		end
			
		if t > 99 then
			image_id = '000'..t
		end

		gt = image.load(opt.gtFolder..image_id..'_10.png')
		local sample_size = 0
		
		label = torch.zeros(gt:size()[2], gt:size()[3], 2)
		valid =  torch.zeros(1, gt:size()[2], gt:size()[3])		

		disparity = image.load(opt.inputFolder..image_id..'/'..opt.disparityName):float()
		d1 = image.load(opt.inputFolder..image_id..'/'..opt.d1Name):float()
		d2 = image.load(opt.inputFolder..image_id..'/'..opt.d2Name):float()
		sample = torch.zeros(2, d1:size()[2], d1:size()[3])
		sample[1] = d1
		sample[2] = d2
		
		patches = torch.zeros(gt:size()[2], gt:size()[3], 2, 15, 15)

		for i=1,gt:size()[2] do
			for j=1,gt:size()[3] do  
				if gt[1][i][j] ~= 0
				then 
					sample_size = sample_size+1
					valid[1][i][j] = 1
					
					if math.abs(disparity[1][i][j] * 255.0 - gt[1][i][j] * 255.0) <= 3
					then
						label[i][j][1] = 1
						label[i][j][2] = 0
					else
						label[i][j][1] = 0
						label[i][j][2] = 1
					end
					patches[i][j] = sample:narrow(2, (i-1)*15 + 1, 15):narrow(3, (i-1)*15 + 1, 15)
 				end
			end
		end
		tot_samples = tot_samples + sample_size

		
		table.insert(dataset, patches)
		table.insert(samples, sample_size)
		table.insert(labels, label)
		table.insert(valids, valid)
		
		print('Maps for '..image_id..' added to dataset!')
	end	
	return dataset, labels, valids, samples, tot_samples
end

