#include "confidence_measures.h"
#include "sgm_confidence_measure.h"
#include "confidence_measures_utility.h"
#include "machine_learning_frameworks.h"

#include "time.h"


/***************************************************************************************************/
/*
* Ensemble
*
/***************************************************************************************************/

//rescaling stereo pair
void ENSEMBLE_rescale(InputArray left_1, InputArray right_1, OutputArray left_2, OutputArray right_2, OutputArray left_3, OutputArray right_3)
{
	printf(" - downscaling stereo pair...\n\n");

	int height_1 = left_1.getMat().rows;
	int width_1 = left_1.getMat().cols;

	int height_2 = height_1 / 2;
	int width_2 = width_1 / 2;

	int height_3 = height_2 / 2;
	int width_3 = width_2 / 2;

	//resize - scale 2
	resize(left_1, left_2, Size(width_1 / 2, height_1 / 2));
	resize(right_1, right_2, Size(width_1 / 2, height_1 / 2));

	//resize - scale 3
	resize(left_1, left_3, Size(width_1 / 4, height_1 / 4));
	resize(right_1, right_3, Size(width_1 / 4, height_1 / 4));
}

//feature vector 
vector<Mat> ENSEMBLE_feature_vector_7(InputArray left_1, InputArray left_2, InputArray left_3, _DSI dsi_1, _DSI dsi_2, _DSI dsi_3)
{
	printf("\n\n ***** Ensemble - feature vector 7 *****\n\n");

	//feature vector 7
	vector<Mat> feature_map;

	Mat T31; // lrc (left right consistency)
	Mat T41, T42, T43; //hg (horizontal gradient)
	Mat T51, T52, T53; //dmv (disparity map variance)

	int height = left_1.getMat().rows;
	int width = left_1.getMat().cols;
	int disp_scale = 256 / dsi_1.num_disp;

	//disparity map L2R
	printf(" - disparity map L2R 1/2/3...\n");
	Mat disparity_L2R_1 = disparity_map_L2R(dsi_1);
	Mat disparity_L2R_2 = disparity_map_L2R(dsi_2);
	Mat disparity_L2R_3 = disparity_map_L2R(dsi_3);

	//disparity map L2R
	printf(" - disparity map R2L...\n");
	Mat disparity_R2L_1 = disparity_map_R2L(dsi_1);

	//left-right consistency - scale 1
	printf(" - left-right consistency - scale 1...\n");
	left_right_consistency_check(disparity_L2R_1, disparity_R2L_1, disp_scale, height, width, T31);

	//horizontal gradient - scale 1/2/3
	printf(" - horizontal gradient - scale 1/2/3...\n");
	horizontal_gradient(left_1, T41);
	horizontal_gradient(left_2, T42);
	horizontal_gradient(left_3, T43);

	//disparity map variance - scale 1/2/3
	printf(" - disparity map variance - scale 1/2/3...\n");
	disparity_variance_measure(disparity_L2R_1, 3, T51);
	disparity_variance_measure(disparity_L2R_2, 3, T52);
	disparity_variance_measure(disparity_L2R_3, 3, T53);


	//resize
	printf(" - upscaling...\n");
	resize(T42, T42, left_1.size(), INTER_LINEAR);
	resize(T43, T43, left_1.size(), INTER_LINEAR);
	resize(T52, T52, left_1.size(), INTER_LINEAR);
	resize(T53, T53, left_1.size(), INTER_LINEAR);

	//feature_vector f7
	feature_map.push_back(T31);
	feature_map.push_back(T41);
	feature_map.push_back(T42);
	feature_map.push_back(T43);
	feature_map.push_back(T51);
	feature_map.push_back(T52);
	feature_map.push_back(T53);

	//return normalize(feature_map);
	return feature_map;
}
vector<Mat> ENSEMBLE_feature_vector_23(InputArray left_1, InputArray left_2, InputArray left_3, InputArray right_1, InputArray right_2, InputArray right_3,
	_DSI dsi_1, _DSI dsi_2, _DSI dsi_3, float pkr_epsilon, float per_s, int zsad_radius, int sge_radius, int b1, int b2)
{
	printf("\n\n ***** Ensemble - feature vector 23  *****\n\n");

	//feature vector 23
	vector<Mat> feature_map;

	//declaration
	Mat c1_1, c1_idx_1, c2_1, c2_idx_1, c_hat_2_1, c_hat_2_idx_1, c_sum_1, NOI_1, c1_R_1, c1_idx_R_1; vector<Mat> local_minima_1;
	Mat c1_2, c1_idx_2, c2_2, c2_idx_2, c_hat_2_2, c_hat_2_idx_2, c_sum_2, NOI_2, c1_R_2, c1_idx_R_2; vector<Mat> local_minima_2;
	Mat c1_3, c1_idx_3, c2_3, c2_idx_3, c_hat_2_3, c_hat_2_idx_3, c_sum_3, NOI_3, c1_R_3, c1_idx_R_3; vector<Mat> local_minima_3;

	//computation of base costs
	printf(" - computation of base costs - scale 1...\n");
	compute_base_costs(dsi_1.values, c1_1, c1_idx_1, c2_1, c2_idx_1, c_hat_2_1, c_hat_2_idx_1, c_sum_1, NOI_1, local_minima_1);
	printf("\n - computation of base costs - scale 2...\n");
	compute_base_costs(dsi_2.values, c1_2, c1_idx_2, c2_2, c2_idx_2, c_hat_2_2, c_hat_2_idx_2, c_sum_2, NOI_2, local_minima_2);
	printf("\n - computation of base costs - scale 3...\n");
	compute_base_costs(dsi_3.values, c1_3, c1_idx_3, c2_3, c2_idx_3, c_hat_2_3, c_hat_2_idx_3, c_sum_3, NOI_3, local_minima_3);

	Mat T01, T02, T03; // pkrn (peak ratio naive)
	Mat T11, T12, T13; // nem (negative entropy)
	Mat T21, T22, T23; // pm (perturbation measure)
	Mat T31; // lrc (left right consistency)
	Mat T41, T42, T43; //hg (horizontal gradient)
	Mat T51, T52, T53; //dmv (disparity map variance)
	Mat T61, T62, T63; //dam (disparity ambiguity measure)
	Mat T71, T72, T73; //zsad (zero mean absolute difference)
	Mat T81; //sge (semi global energy)


	int height_1 = left_1.getMat().rows;
	int width_1 = left_1.getMat().cols;

	int height_2 = left_2.getMat().rows;
	int width_2 = left_2.getMat().cols;

	int height_3 = left_3.getMat().rows;
	int width_3 = left_3.getMat().cols;

	int disp_scale = 256 / dsi_1.num_disp;

	//disparity map L2R
	printf("\n - disparity map L2R 1/2/3...\n");
	Mat disparity_L2R_1 = disparity_map_L2R(dsi_1);
	Mat disparity_L2R_2 = disparity_map_L2R(dsi_2);
	Mat disparity_L2R_3 = disparity_map_L2R(dsi_3);

	//disparity map R2L
	printf(" - disparity map R2L...\n");
	Mat disparity_R2L_1 = disparity_map_R2L(dsi_1);

	//peak ratio naive - scale 1/2/3
	printf(" - peak ratio naive - scale 1/2/3...\n");
	peak_ratio(c1_1, c2_1, pkr_epsilon, height_1, width_1, T01);
	peak_ratio(c1_2, c2_2, pkr_epsilon, height_2, width_2, T02);
	peak_ratio(c1_3, c2_3, pkr_epsilon, height_3, width_3, T03);

	//negative entropy - scale 1/2/3
	printf(" - negative entropy - scale 1/2/3...\n");
	negative_entropy_measure(dsi_1.values, height_1, width_1, T11);
	negative_entropy_measure(dsi_2.values, height_2, width_2, T12);
	negative_entropy_measure(dsi_3.values, height_3, width_3, T13);

	//perturbation measure - scale 1/2/3
	printf(" - perturbation measure - scale 1/2/3...\n");
	perturbation_measure(dsi_1.values, c1_1, c1_idx_1, per_s, height_1, width_1, T21);
	perturbation_measure(dsi_2.values, c1_2, c1_idx_2, per_s, height_2, width_2, T22);
	perturbation_measure(dsi_3.values, c1_3, c1_idx_3, per_s, height_3, width_3, T23);

	//left-right consistency - scale 1
	printf(" - left-right consistency - scale 1...\n");
	left_right_consistency_check(disparity_L2R_1, disparity_R2L_1, disp_scale, height_1, width_1, T31);

	//horizontal gradient - scale 1/2/3
	printf(" - horizontal gradient - scale 1/2/3...\n");
	horizontal_gradient(left_1, T41);
	horizontal_gradient(left_2, T42);
	horizontal_gradient(left_3, T43);

	//disparity map variance - scale 1/2/3
	printf(" - disparity map variance - scale 1/2/3...\n");
	disparity_variance_measure(disparity_L2R_1, 3, T51);
	disparity_variance_measure(disparity_L2R_2, 3, T52);
	disparity_variance_measure(disparity_L2R_3, 3, T53);

	//disparity ambiguity measure - scale 1/2/3
	printf(" - disparity ambiguity measure - scale 1/2/3...\n");
	disparity_ambiguity_measure(c1_idx_1, c2_idx_1, height_1, width_1, T61);
	disparity_ambiguity_measure(c1_idx_2, c2_idx_2, height_2, width_2, T62);
	disparity_ambiguity_measure(c1_idx_3, c2_idx_3, height_3, width_3, T63);

	//zero mean absolute difference - scale 1/2/3
	printf(" - zero mean absolute difference - scale 1/2/3...\n");
	zero_mean_sum_of_absolute_differences(left_1, right_1, c1_1, c1_idx_1, zsad_radius, height_1, width_1, T71);
	zero_mean_sum_of_absolute_differences(left_2, right_2, c1_2, c1_idx_2, zsad_radius, height_2, width_2, T72);
	zero_mean_sum_of_absolute_differences(left_3, right_3, c1_3, c1_idx_3, zsad_radius, height_3, width_3, T73);

	//semiglobal energy
	printf(" - semiglobal energy...\n");
	SGM_semi_global_energy(c1_1, disparity_L2R_1, b1, b2, sge_radius, height_1, width_1, T81);

	//resize
	//printf(" - upscaling 02...\n");
	resize(T02, T02, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 03...\n");
	resize(T03, T03, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 12...\n");
	resize(T12, T12, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 13...\n");
	resize(T13, T13, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 22...\n");
	resize(T22, T22, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 23...\n");
	resize(T23, T23, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 42...\n");
	resize(T42, T42, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 43...\n");
	resize(T43, T43, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 52...\n");
	resize(T52, T52, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 53...\n");
	resize(T53, T53, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 62...\n");
	resize(T62, T62, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 63...\n");
	resize(T63, T63, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 72...\n");
	resize(T72, T72, left_1.size(), INTER_LINEAR);
	//printf(" - upscaling 73...\n");
	resize(T73, T73, left_1.size(), INTER_LINEAR);

	//feature_vector f23
	feature_map.push_back(T01);
	feature_map.push_back(T02);
	feature_map.push_back(T03);
	feature_map.push_back(T11);
	feature_map.push_back(T12);
	feature_map.push_back(T13);
	feature_map.push_back(T21);
	feature_map.push_back(T22);
	feature_map.push_back(T23);
	feature_map.push_back(T31);
	feature_map.push_back(T41);
	feature_map.push_back(T42);
	feature_map.push_back(T43);
	feature_map.push_back(T51);
	feature_map.push_back(T52);
	feature_map.push_back(T53);
	feature_map.push_back(T61);
	feature_map.push_back(T62);
	feature_map.push_back(T63);
	feature_map.push_back(T71);
	feature_map.push_back(T72);
	feature_map.push_back(T73);
	feature_map.push_back(T81);
	//printf("Features pushed back\n");

	//return normalize(feature_map);
	return feature_map;
}

//train data for random forest 
void ENSEMBLE_write_train_sample(vector<Mat> feature_vector, Mat disparity_map, Mat GT, int disp_scale, float threshold, char * file_path, int invalid)
{
	printf("\n\n ***** Ensemble - write_train_sample *****\n\n");

	FILE* training_set = fopen(file_path, "a" );

	for (int i = 0; i < disparity_map.rows; i++)
		for (int j = 0; j < disparity_map.cols; j++)
		{
			if (invalid == 0)
			{
				if (GT.at<ushort>(i, j) != 0)
				{
					for (int f = 0; f < feature_vector.size(); f++)
						fprintf(training_set, "%f,", feature_vector[f].at<float>(i,j));
					if (fabs(disparity_map.at<float>(i, j) - GT.at<ushort>(i, j)/256.0) <= 3)
						fprintf(training_set, "%f\n", 1.0);
					else
						fprintf(training_set, "%f\n", 0.0);
				}
			}
			else
			{
				if (GT.at<uchar>(i, j) != 255)
				{
					for (int f = 0; f < feature_vector.size(); f++)
						fprintf(training_set, "%f,", feature_vector[f].at<float>(i,j));
					if (fabs(disparity_map.at<float>(i, j) - GT.at<uchar>(i, j)) <= 3)
						fprintf(training_set, "%f\n", 1.0);
					else
						fprintf(training_set, "%f\n", 0.0);
				}
			}
		}
	fclose(training_set);	
}

/***************************************************************************************************/
/*
* GCP
*
/***************************************************************************************************/

// 1. write train sample to file
void GCP_write_train_sample(InputArray image_L, InputArray image_R, Mat GT, _DSI DSI_LR, float epsilon_PKR, float sigma_MLM, float sigma_AML, char *output_path, int invalid)
{
	//width, height
	int height = DSI_LR.height;
	int width = DSI_LR.width;
	int d_min = DSI_LR.d_min;
	int d_max = DSI_LR.d_max;
	int num_disp = DSI_LR.num_disp;
	int disp_scale = 256 / num_disp;
	vector<Mat> values = DSI_LR.values;

	//compute DD info 
	int edgeThresh = 1;
	int lowThreshold;
	int const max_lowThreshold = 50;
	int ratio = 3;
	int kernel_size = 3;

	//generate right dsi.
	printf("generate right dsi...\n");
	_DSI DSI_RL = DSI_left2right(DSI_LR);

	//compute left and right disparity maps
	printf("compute left and right disparity maps...\n");
	Mat disparity_L2R = disparity_map_L2R(DSI_LR);
	Mat disparity_R2L = disparity_map_R2L(DSI_LR);

	printf("compute c_1, c_2, c^_2 in the paper as well as sum of matching costs...\n");
	Mat c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, c1_R, c1_idx_R;
	vector<Mat> local_minima;
	compute_base_costs(DSI_LR.values, c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, local_minima);
	//compute minimum right to left
	printf("compute minimum right to left...\n");
	minimum(DSI_RL.values, c1_R);

	//output
	Mat confidence_map, normalized, equalized, to_visualize_map;
	Mat GCP[8];
	for (int f = 0; f < 8; f ++)
		GCP[f].create(Size(width, height), CV_32F);

	matching_score_measure(c1, height, width, GCP[0]);
	distance_to_border(disparity_L2R,  GCP[1]);	
	maximum_margin(c1, c_hat_2, height, width, GCP[2]);
	attainable_maximum_likelihood(values, c1, sigma_AML, height, width, GCP[3]);
	left_right_consistency_check(disparity_L2R, disparity_R2L, disp_scale, height, width, GCP[4]);
	left_right_difference(c1, c2, c1_R, disparity_L2R, disp_scale, 0.1, height, width, GCP[5]);
	compute_DD(disparity_L2R, 50, 3, 1, GCP[6]);
	difference_with_median(disparity_L2R, 5, height, width, GCP[7]);
	printf("GCP features ready!\n");

	FILE* training_set = fopen(output_path, "a" );

	for (int i = 0; i < height; i++)
		for (int j = 0; j < width; j++)
		{
			if (invalid == 0)
			{
				if (GT.at<ushort>(i, j) != 0)
				{
					for (int f = 0; f < 8; f++)
						fprintf(training_set, "%f,", GCP[f].at<float>(i,j));
					if (fabs(disparity_L2R.at<float>(i, j) - GT.at<ushort>(i, j)/256.0) <= 3)
						fprintf(training_set, "%f\n", 1.0);
					else
						fprintf(training_set, "%f\n", 0.0);
				}
			}
			else
			{
				if (GT.at<uchar>(i, j) != 255)
				{
					for (int f = 0; f < 8; f++)
						fprintf(training_set, "%f,", GCP[f].at<float>(i,j));
					if (fabs(disparity_L2R.at<float>(i, j) - GT.at<uchar>(i, j)) <= 3)
						fprintf(training_set, "%f\n", 1.0);
					else
						fprintf(training_set, "%f\n", 0.0);
				}
			}
		}
	fclose(training_set);	
				
	return;
}

// 2. predict confidence map
void GCP_confidence(InputArray image_L, InputArray image_R, _DSI DSI_LR, float epsilon_PKR, float sigma_MLM, float sigma_AML, char *forest_path, OutputArray confidence_map)
{
	CvRTrees forest;
	forest.load(forest_path);
	
	//width, height
	int height = DSI_LR.height;
	int width = DSI_LR.width;
	int d_min = DSI_LR.d_min;
	int d_max = DSI_LR.d_max;
	int num_disp = DSI_LR.num_disp;
	int disp_scale = 256 / num_disp;
	vector<Mat> values = DSI_LR.values;

	//compute DD info 
	int edgeThresh = 1;
	int lowThreshold;
	int const max_lowThreshold = 50;
	int ratio = 3;
	int kernel_size = 3;

	//generate right dsi.
	printf("generate right dsi...\n");
	_DSI DSI_RL = DSI_left2right(DSI_LR);

	//compute left and right disparity maps
	printf("compute left and right disparity maps...\n");
	Mat disparity_L2R = disparity_map_L2R(DSI_LR);
	Mat disparity_R2L = disparity_map_R2L(DSI_LR);

	printf("compute c_1, c_2, c^_2 in the paper as well as sum of matching costs...\n");
	Mat c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, c1_R, c1_idx_R, variance;
	vector<Mat> local_minima;
	compute_base_costs(DSI_LR.values, c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, local_minima);

	//compute minimum right to left
	printf("compute minimum right to left...\n");
	minimum(DSI_RL.values, c1_R);

	//output
	confidence_map.create(height, width, CV_32F);
	Mat _confidence_map = confidence_map.getMat();

	Mat GCP[8];
	for (int f = 0; f < 8; f ++)
		GCP[f].create(Size(width, height), CV_32F);
	matching_score_measure(c1, height, width, GCP[0]);
	distance_to_border(disparity_L2R,  GCP[1]);	
	maximum_margin(c1, c_hat_2, height, width, GCP[2]);
	attainable_maximum_likelihood(values, c1, sigma_AML, height, width, GCP[3]);
	left_right_consistency_check(disparity_L2R, disparity_R2L, disp_scale, height, width, GCP[4]);
	left_right_difference(c1, c2, c1_R, disparity_L2R, disp_scale, 0.1, height, width, GCP[5]);
	compute_DD(disparity_L2R, 50, 3, 1, GCP[6]);
	difference_with_median(disparity_L2R, 5, height, width, GCP[7]);
	printf("GCP features ready!\n");

	Mat test_sample;
	test_sample.create(1, 8, CV_32FC1);


	for (int i = 0; i < height; i++)
		for (int j = 0; j < width; j++)
		{
			for (int f = 0; f < 8; f++)
				test_sample.at<float>(0,f) = GCP[f].at<float>(i,j);
			_confidence_map.at<float>(i,j) = forest.predict(test_sample, Mat());

		}

	return;
}

/***************************************************************************************************/
/*
* Park
*
/***************************************************************************************************/

// 1. write train sample to file
void Park_write_train_sample(InputArray image_L, InputArray image_R, Mat GT, _DSI DSI_LR, float epsilon_PKR, float sigma_MLM, float sigma_AML, float s_PER, char *output_path, int invalid)
{
	//width, height
	int height = DSI_LR.height;
	int width = DSI_LR.width;
	int d_min = DSI_LR.d_min;
	int d_max = DSI_LR.d_max;
	int num_disp = DSI_LR.num_disp;
	int disp_scale = 256 / num_disp;
	vector<Mat> values = DSI_LR.values;

	//compute DD info 
	int edgeThresh = 1;
	int lowThreshold;
	int const max_lowThreshold = 50;
	int ratio = 3;
	int kernel_size = 3;

	//generate right dsi.
	printf("generate right dsi...\n");
	_DSI DSI_RL = DSI_left2right(DSI_LR);

	//compute left and right disparity maps
	printf("compute left and right disparity maps...\n");
	Mat disparity_L2R = disparity_map_L2R(DSI_LR);
	Mat disparity_R2L = disparity_map_R2L(DSI_LR);

	printf("compute c_1, c_2, c^_2 in the paper as well as sum of matching costs...\n");
	Mat c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, c1_R, c1_idx_R, variance;
	vector<Mat> local_minima;
	compute_base_costs(DSI_LR.values, c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, local_minima);

	//compute minimum right to left
	printf("compute minimum right to left...\n");
	minimum(DSI_RL.values, c1_R);

	//output
	Mat confidence_map, normalized, equalized, to_visualize_map;
	Mat Park[22];
	for (int f = 0; f < 22; f ++)
		Park[f].create(Size(width, height), CV_32F);

	peak_ratio(c1, c_hat_2, epsilon_PKR, height, width, Park[0]);
	peak_ratio(c1, c2, epsilon_PKR, height, width, Park[1]);
	matching_score_measure(c1, height, width, Park[2]);
	maximum_margin(c1, c2, height, width, Park[3]);
	winner_margin(c1, c_hat_2, c_sum, height, width, Park[4]);
	maximum_likelihood_measure(values, c1, sigma_MLM, height, width, Park[5]);
	perturbation_measure(values, c1, c1_idx, s_PER, height, width, Park[6]);
	negative_entropy_measure(values, height, width, Park[7]);
	left_right_difference(c1, c2, c1_R, disparity_L2R, disp_scale, 0.1, height, width, Park[8]);
	curvature(DSI_LR.values, c1_idx, height, width, Park[9]);

	var_measure(disparity_L2R, 2, Park[10]);
	var_measure(disparity_L2R, 3, Park[11]);
	var_measure(disparity_L2R, 4, Park[12]);
	var_measure(disparity_L2R, 5, Park[13]);

	compute_DD(disparity_L2R, max_lowThreshold, ratio, kernel_size, Park[14]);
	difference_with_median(disparity_L2R, 5, height, width, Park[15]);
	difference_with_median(disparity_L2R, 7, height, width, Park[16]);
	difference_with_median(disparity_L2R, 9, height, width, Park[17]);
	difference_with_median(disparity_L2R, 11, height, width, Park[18]);

	left_right_consistency_check(disparity_L2R, disparity_R2L, disp_scale, height, width, Park[19]);
	horizontal_gradient(image_L, Park[20]);
	distance_to_left_border(disparity_L2R, d_max, Park[21]);
	Park[21] = Park[21] * -1;
	printf("Park features ready!\n");

	FILE* training_set = fopen(output_path, "a" );

	for (int i = 0; i < height; i++)
		for (int j = 0; j < width; j++)
		{
			if (invalid == 0)
			{
				if (GT.at<ushort>(i, j) != 0)
				{
					for (int f = 0; f < 22; f++)
						fprintf(training_set, "%f,", Park[f].at<float>(i,j));
					if (fabs(disparity_L2R.at<float>(i, j) - GT.at<ushort>(i, j)/256.0) <= 3)
						fprintf(training_set, "%f\n", 1.0);
					else
						fprintf(training_set, "%f\n", 0.0);
				}
			}
			else
			{
				if (GT.at<uchar>(i, j) != 255)
				{
					for (int f = 0; f < 22; f++)
						fprintf(training_set, "%f,", Park[f].at<float>(i,j));
					if (fabs(disparity_L2R.at<float>(i, j) - GT.at<uchar>(i, j)) <= 3)
						fprintf(training_set, "%f\n", 1.0);
					else
						fprintf(training_set, "%f\n", 0.0);
				}
			}
		}
	fclose(training_set);	
				
	return;
}

// 2. predict confidence map
void Park_confidence(InputArray image_L, InputArray image_R, _DSI DSI_LR, float epsilon_PKR, float sigma_MLM, float sigma_AML, float s_PER, char *forest_path, OutputArray confidence_map)
{
	CvRTrees forest;
	forest.load(forest_path);

	//width, height
	int height = DSI_LR.height;
	int width = DSI_LR.width;
	int d_min = DSI_LR.d_min;
	int d_max = DSI_LR.d_max;
	int num_disp = DSI_LR.num_disp;
	int disp_scale = 256 / num_disp;
	vector<Mat> values = DSI_LR.values;

	//compute DD info 
	int edgeThresh = 1;
	int lowThreshold;
	int const max_lowThreshold = 50;
	int ratio = 3;
	int kernel_size = 3;

	//generate right dsi.
	printf("generate right dsi...\n");
	_DSI DSI_RL = DSI_left2right(DSI_LR);

	//compute left and right disparity maps
	printf("compute left and right disparity maps...\n");
	Mat disparity_L2R = disparity_map_L2R(DSI_LR);
	Mat disparity_R2L = disparity_map_R2L(DSI_LR);

	printf("compute c_1, c_2, c^_2 in the paper as well as sum of matching costs...\n");
	Mat c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, c1_R, c1_idx_R, variance;
	vector<Mat> local_minima;
	compute_base_costs(DSI_LR.values, c1, c1_idx, c2, c2_idx, c_hat_2, c_hat_2_idx, c_sum, NOI, local_minima);

	//compute minimum right to left
	printf("compute minimum right to left...\n");
	minimum(DSI_RL.values, c1_R);

	//output
	confidence_map.create(height, width, CV_32F);
	Mat _confidence_map = confidence_map.getMat();

	Mat Park[22];
	for (int f = 0; f < 22; f ++)
		Park[f].create(Size(width, height), CV_32F);
	peak_ratio(c1, c_hat_2, epsilon_PKR, height, width, Park[0]);
	peak_ratio(c1, c2, epsilon_PKR, height, width, Park[1]);
	matching_score_measure(c1, height, width, Park[2]);
	maximum_margin(c1, c2, height, width, Park[3]);
	winner_margin(c1, c_hat_2, c_sum, height, width, Park[4]);
	maximum_likelihood_measure(values, c1, sigma_MLM, height, width, Park[5]);
	perturbation_measure(values, c1, c1_idx, s_PER, height, width, Park[6]);
	negative_entropy_measure(values, height, width, Park[7]);
	left_right_difference(c1, c2, c1_R, disparity_L2R, disp_scale, 0.1, height, width, Park[8]);
	curvature(DSI_LR.values, c1_idx, height, width, Park[9]);

	// VAR
	var_measure(disparity_L2R, 2, Park[10]);
	var_measure(disparity_L2R, 3, Park[11]);
	var_measure(disparity_L2R, 4, Park[12]);
	var_measure(disparity_L2R, 5, Park[13]);

	// distance to edge
	compute_DD(disparity_L2R, max_lowThreshold, ratio, kernel_size, Park[14]);

	// MED
	difference_with_median(disparity_L2R, 5, height, width, Park[15]);
	difference_with_median(disparity_L2R, 7, height, width, Park[16]);
	difference_with_median(disparity_L2R, 9, height, width, Park[17]);
	difference_with_median(disparity_L2R, 11, height, width, Park[18]);

	left_right_consistency_check(disparity_L2R, disparity_R2L, disp_scale, height, width, Park[19]);
	horizontal_gradient(image_L, Park[20]);
	distance_to_left_border(disparity_L2R, d_max, Park[21]);
	Park[21] = Park[21] * -1;	
	printf("Park features ready!\n");

	Mat test_sample;
	test_sample.create(1, 22, CV_32FC1);


	for (int i = 0; i < height; i++)
		for (int j = 0; j < width; j++)
		{
			for (int f = 0; f < 22; f++)
				test_sample.at<float>(0,f) = Park[f].at<float>(i,j);
			_confidence_map.at<float>(i,j) = forest.predict(test_sample, Mat());
		}
	return;
}
